; -----------------------------------------------------------------------------
; "The War of the Worlds" - A dynamic tape loader e-book for fast readers
; -----------------------------------------------------------------------------
; Einar Saukas - WOOT 2023
; -----------------------------------------------------------------------------

SA_LD_RET       EQU $053f
LD_BLOCK        EQU $0802
MSG_BYTES       EQU $09ec
CHAN_OPEN       EQU $1601
CHARS           EQU $5c36
ATTR_P          EQU $5c8d


org 60000
        ; move this code to address 60000 in case it was loaded elsewhere
code_begin:
        ld h,b
        ld l,c
        ld de,code_begin
        ld bc,code_end-code_begin
        ldir
        jp code_start

; -----------------------------------------------------------------------------
; Main routine: execute tape loader and handle errors.
; -----------------------------------------------------------------------------
code_start:
        ; load header
        ld ix,$ff00                     ; initial address
        ld de,17                        ; size
        xor a                           ; type HEADER
        scf                             ; mode LOAD
        call LD_BLOCK                   ; load with error handling
        ; open channel 2
        ld a,2
        call CHAN_OPEN
        ; print "Bytes:"
        ld hl,MSG_BYTES
        ld b,7
loop_bytes:
        ld a,(hl)
        inc hl
        rst $10
        djnz loop_bytes
        ; print space
        ld a,32
        rst $10
        ; print block name
        ld hl,$ff01
        ld b,10
loop_name:
        ld a,(hl)
        inc hl
        rst $10
        djnz loop_name
        ; preserve registers
        di
        exx
        push hl
        push iy
        ; extended load data block
        ld bc,$ff00                     ; circular input buffer address
        call ld_bytes
        ; restore registers
        pop iy
        pop hl
        exx
        ; error handling
        call SA_LD_RET
        jp LD_BLOCK+3

; -----------------------------------------------------------------------------
; Tape loader: search for leader sign and sync sign, then load data from tape.
; -----------------------------------------------------------------------------
ld_bytes:
        ; set IX = BC'
        push bc
        pop ix
        exx
        ; set A'=255 (type DATA), set C flag, reset Z flag
        xor a
        sub 1
        ex af,af'
        ; initialize font address
        ld hl,(CHARS)
        ld (charset1+1),hl
        ; initialize initial attribute address
        ld hl,$5800
        ld (attr_pos1+1),hl
        ; initialize text attribute
        ld hl,ATTR_P
        ld a,(hl)
        ld (show1+1),a
        ; initialize hidden attribute
        rrca
        rrca
        rrca
        xor (hl)
        and $07
        xor (hl)
        ld (hide1+1),a
        ; initialize cursor attribute
        ld a,(hl)
        rlca
        rlca
        rlca
        xor (hl)
        and $38
        xor (hl)
        ld (cursor1+1),a
        ; terminator counter
        ld de,3
        ; initial state
        ld iy,st_main
        ; set white border
        ld a,$0f
        out ($fe),a
        ; read EAR into C
        in a,($fe)
        rra
        and $20
        or $02                          ; red border
        ld c,a
        cp a                            ; reset flags
        ; start loading from tape
ld_break:
        ret nz                          ; return if BREAK
ld_start:
        ; wait for single pulse
        call ld_edge1
        jr nc,ld_break                  ; restart loading if no pulse
        ; wait about 1 second
        ld hl,$0415
ld_wait:
        djnz ld_wait
        dec hl
        ld a,h
        or l
        jr nz,ld_wait
        ; check double pulse
        call ld_edge2
        jr nc,ld_break                  ; restart loading if no double pulse
        ; search for leader sign (at least 256 double pulses)
ld_leader:
        ; check duration of next double pulse
        ld b,$9c
        call ld_edge2
        jr nc,ld_break                  ; restart loading if no double pulse
        ld a,$c6
        cp b
        jr nc,ld_start                  ; restart loading if incorrect duration
        inc h
        jr nz,ld_leader                 ; repeat 256 times
        ; keep checking single pulses until sync sign is found
ld_sync:
        ld b,$c9
        call ld_edge1                   ; check next single pulse
        jr nc,ld_break                  ; restart loading if no pulse
        ld a,b
        cp $d4
        jr nc,ld_sync                   ; loop until short sync pulse is found
        call ld_edge1                   ; check 2nd short sync pulse
        ret nc                          ; error if no short sync pulse
        ; sync validated, now it's time to load actual data
        ld a,c
        xor $03                         ; switch border color
        ld c,a
        ld h,$00                        ; initialize checksum
        ld b,$b0                        ; next time delay
        jr ld_marker                    ; load first byte

; -----------------------------------------------------------------------------
; Validate flag byte loaded from tape (L) against expected value (A).
; -----------------------------------------------------------------------------
ld_flag:
        rl c                            ; preserve carry flag
        xor l                           ; compare expected and loaded flag byte
        ret nz                          ; error if different
        ld a,c
        rra                             ; restore carry flag
        ld c,a
        inc de                          ; extra delay (6T)
        dec de                          ; extra delay (6T)
        ex af,af'                       ; save flags
        jr ld_next                      ; load next byte

; -----------------------------------------------------------------------------
; Main tape loader loop: load actual data until terminator byte 255 is found.
; -----------------------------------------------------------------------------
ld_loop:
        ex af,af'                       ; restore flags
        jr nz,ld_flag                   ; if 1st byte then validate "flag byte"
        ex af,af'                       ; save flags
        ; put loaded data into circular buffer
        ld (ix+0),l
        inc ixl
        ; increment register E unless terminator value 255 was loaded from tape
        ld a,$fe
        cp l
        sbc a,a
        inc a
        add a,e
        ld e,a
ld_next:
        ld b,$b2                        ; next time delay
ld_marker:
        ; load byte (8 bits)
        ld l,1                          ; store end marker (8x) into register L
ld_byte:
        call ld_edge2                   ; count duration of next double pulse
        ret nc                          ; return if error
        ld a,$cb                        ; identify bit 0 or 1 from duration
        cp b
        rl l                            ; store new bit into register L
        ld b,$b0                        ; next time delay
        jp nc,ld_byte                   ; repeat 8 times (1 byte)
        ; update checksum
        ld a,h
        xor l
        ld h,a
        nop                             ; extra delay (4T)
        ; repeat until terminator is loaded from tape
        dec e
        jr nz,ld_loop
        ; set carry if checksum is correct
        ld a,h
        cp $01
        ret

; -----------------------------------------------------------------------------
; Count duration of next double pulse. This is identical to original ROM loader
; routine except it runs one step of the state machine before each tape pulse.
; At the end of each execution, this routine will reset carry flag in case of 
; error (resetting zero flag to indicate BREAK, setting to indicate timeout).
; -----------------------------------------------------------------------------
ld_edge2:
        call ld_edge1
        ret nc                          ; return if error
ld_edge1:
        ; original delay (before pulse detection) spent 354T doing nothing:
;       ld a,22
;delay: dec a
;       jr nz,delay
        ; new delay (before pulse detection) spends 354T running state machine:
        exx                             ; preserve main registers
        jp (iy)                         ; execute current state
exit:
        exx                             ; restore main registers
delay:
        dec a                           ; remaining delay (A*16T)-5T
        jr nz,delay
        ; now start counting time interval until next edge type change
        and a                           ; reset carry
ld_sample:
        inc b                           ; duration counter
        ret z                           ; time expired?
        ld a,$7f
        in a,($fe)                      ; read port $7ffe
        rra
        ret nc                          ; return if BREAK
        ; check if pulse signal changed
        xor c
        and $20
        jr z,ld_sample                  ; keep checking until edge type changed
        ; switch edge type
        ld a,c
        cpl
        ld c,a
        ; alternate border color
        and $07
        or $08
        out ($fe),a
        scf                             ; indicate success
        ret

; -----------------------------------------------------------------------------
; State "char": Draw character at current position (one pixel line at a time).
; -----------------------------------------------------------------------------
st_char:
        ; copy next byte from charset table to screen
        ld a,(de)                       ; read character byte
        ld (hl),a                       ; copy to screen
        inc de                          ; next character byte position
        inc h                           ; next screen position
        ld a,h
        and 7
        jr z,char_end                   ; copied all 8 bytes?
        ld a,(bc)                       ; extra delay (7T)
        jr char_cont
char_end:
        ld iy,st_show                   ; next state st_show (if finished copy)
char_cont:
        ld a,(bc)                       ; extra delay (7T)
        ld a,16                         ; remaining delay
        jr exit                         ; exit

; -----------------------------------------------------------------------------
; State "erase": Erase character under current position and advance current
; position. It's executed only once, before drawing cursor at new position.
; -----------------------------------------------------------------------------
st_erase:
        ; increment attribute position
        inc hl
        ld a,h
        cp $5b
        sbc a,a
        and h
        or $58
        ld h,a
        ld (attr_pos1+1),hl
        ; calculate attribute position below
        ld de,31
        add hl,de
        ld a,h
        cp $5b
        sbc a,a
        and h
        or $58
        ld h,a
hide1:
        ; erase content under current position
        ld (hl),63
        ld iy,st_cursor                 ; next state st_cursor
        ret c                           ; extra delay (5T)
        ld a,12                         ; remaining delay
exit12:
        jr exit                         ; exit

; -----------------------------------------------------------------------------
; State "show": Make current position visible.
; -----------------------------------------------------------------------------
st_show:
        ld hl,(attr_pos1+1)             ; current attribute position
show1:
        ld (hl),56                      ; show content at current position
        ld iy,st_erase                  ; next state st_erase
        ld a,17                         ; remaining delay
        jr exit12                       ; exit

; -----------------------------------------------------------------------------
; State "enter": Erase character under current position and advance current
; position. Keep doing it until it reaches column zero.
; -----------------------------------------------------------------------------
st_enter:
        ld hl,(attr_pos1+1)             ; current attribute position
        ; increment attribute position
        inc hl
        ld a,h
        cp $5b
        sbc a,a
        and h
        or $58
        ld h,a
        ld (attr_pos1+1),hl
        ; calculate attribute position below
        ld de,31
        add hl,de
        ld a,h
        cp $5b
        sbc a,a
        and h
        or $58
        ld h,a
        ; erase content under current position
        ld a,(hide1+1)
        ld (hl),a
        ; check current column
        ld a,l
        inc a
        and e
        jr z,column_zero                ; column zero?
        ld a,(bc)                       ; extra delay (7T)
        jr column_notzero
column_zero:
        ld iy,st_main                   ; next state st_main (if column zero)
column_notzero:
        ld d,255                        ; D=255 to indicate inactive cursor
        ld a,8                          ; remaining delay
        jr exit12                       ; exit

; -----------------------------------------------------------------------------
; State "main": Check next byte from input buffer and decide what to do.
; -----------------------------------------------------------------------------
st_main:
        ; anything on input buffer?
        ld a,ixl
        cp c
        jr z,empty
        ; get next byte from input buffer
        ld a,(bc)
        inc c
        ; check ENTER
        cp 13
        jr z,newline
        ; check SPACE
        cp 32
        jr z,space
        ; check non-ASCII character (char>127)
        add a,a                         ; char*2
        jr c,illegal
        ; process character
        ld h,0
        ld l,a
        add hl,hl                       ; char*4
        add hl,hl                       ; char*8
charset1:
        ld de,$3C00
        add hl,de
        ex de,hl                        ; charset table address
        ld hl,(attr_pos1+1)             ; current attribute position
        ; convert to current pixel screen position
        ld a,h
        add a,a
        add a,a
        add a,a
        and $58
        ld h,a
        ld iy,st_char                   ; next state st_char
        ld a,8                          ; remaining delay
        jr exit12                       ; exit
illegal:
        ld iy,st_cursor                 ; next state st_cursor
        ld a,14                         ; remaining delay
        jr exit12                       ; exit
empty:
        ld a,18                         ; remaining delay
        jr exit12                       ; exit
space:
        ; erase cursor at current position
        ld hl,(attr_pos1+1)             ; current attribute position
        ld a,(hide1+1)
        ld (hl),a
        ld iy,st_erase                  ; next state st_erase
        ret nz                          ; extra delay (5T)
        ld a,13                         ; remaining delay
exit10:
        jp exit                         ; exit
newline:
        ; erase cursor at current position
        ld hl,(attr_pos1+1)             ; current attribute position
        ld a,(hide1+1)
        ld (hl),a
        ; check column zero and active cursor
        ld a,l
        and 31
        or d
        jr nz,erase_line                ; erase remaining line, unless it was
                                        ; at column zero with active cursor
        ld d,255                        ; D=255 to indicate inactive cursor
        jr finished
erase_line:
        ld iy,st_enter                  ; next state st_enter
finished:
        ld a,11                         ; remaining delay
        jr exit22                       ; exit

; -----------------------------------------------------------------------------
; State "cursor": Draw cursor at current position.
; -----------------------------------------------------------------------------
st_cursor:
attr_pos1:
        ld hl,$5800                     ; current attribute position
cursor1:
        ld (hl),0                       ; set cursor at current position
        ld iy,st_main                   ; next state st_main
        xor a
        ld d,a                          ; D=0 to indicate active cursor
        ld a,17                         ; remaining delay
exit22:
        jr exit10                       ; exit

; -----------------------------------------------------------------------------
code_end:
